"""Basic virtual IOTile device class for testing other components interactions with IOTile devices"""

from iotile.core.hw.virtual.virtualdevice_simple import SimpleVirtualDevice


class SimpleVirtualDevice(SimpleVirtualDevice):
    """Mock IOTileDevice that allows capturing interactions and injecting faults

    Args:
        args (dict): Any arguments that you want to pass to create this device.
            None are supported.
    """

    def __init__(self, args):
        super(SimpleVirtualDevice, self).__init__(1, 'Simple')



"""Virtual device that acts like a blank tile with just an executive installed"""

from iotile.core.hw.virtual.virtualdevice_simple import SimpleVirtualDevice


class NoAppVirtualDevice(SimpleVirtualDevice):
    """Virtual device that acts like a blank tile with just an executive loaded

    Args:
        args (dict): Any arguments that you want to pass to create this device.
            None are supported.
    """

    def __init__(self, args):
        super(NoAppVirtualDevice, self).__init__(1, 'NO APP')



"""A Mock class for overriding open calls."""

import pytest

from contextlib import contextmanager
from io import StringIO


class MockFileOpen:
    def __init__(self):
        """Constructor."""

        self.files = {}
        self._real_open = open

    @contextmanager
    def mock_open(self, filename, params):
        """Mock open a file."""

        if filename not in self.files:
            with self._real_open(filename, params) as realfile:
                yield realfile
        else:
            if params[0] == 'r':
                self.files[filename].seek(0)
            elif params[0] == 'w':
                self.files[filename] = StringIO()

            yield self.files[filename]

    def add_file(self, path, contents=""):
        """Add a mock file with the given contents.

        Args:
            path (string): The path to the fake file
            contents (string): Optional file contents
        """

        self.files[path] = StringIO(contents)

    def from_file(self, path, content_path):
        """Add a mock file from the actual file.

        Args:
            path (string): The path to the fake file
            content_path (string): The actual file path
        """

        with self._real_open(content_path, "rb") as f:
            contents = f.read()

        self.add_file(path, contents)


@pytest.fixture
def mock_fs():
    """Create a mock fs object that can be used to patch file opens."""

    mock = MockFileOpen()
    return mock



"""Basic virtual IOTile device class for testing other components interactions with IOTile devices"""

import random
from iotile.core.hw.virtual.virtualdevice_simple import SimpleVirtualDevice, rpc


class SensorGraphTestDevice(SimpleVirtualDevice):
    """Mock IOTileDevice that can be used to test sensor graphs.

    Args:
        args (dict): Any arguments that you want to pass to create this device.
            None are supported.
    """

    def __init__(self, args):
        super(SensorGraphTestDevice, self).__init__(1, 'Simple')

    @rpc(8, 0x8000, "", "L")
    def random_int(self):
        """Return a random int in the range [0, 100)."""

        return [random.randrange(0, 100)]

    @rpc(11, 0x8001, "", "L")
    def return_fixed_int(self):
        """Return the fixed intger 42."""

        return [42]



"""Reference device for testing the individual report format
"""

from iotile.core.hw.exceptions import DevicePushError
from iotile.core.hw.virtual import SimpleVirtualDevice
from iotile.core.hw.reports import IndividualReadingReport, IOTileReading, BroadcastReport


class RealtimeTestDevice(SimpleVirtualDevice):
    """Mock IOTileDevice that streams and traces data periodically

    This device can be configured to stream data on any streams at any interval.
    It can be used for testing realtime streaming functionality of any other portion
    of the IOTile stack.

    If no other arguments are passed, this device defaults to producing the value 100
    on stream 0x1001 every second.  If a streams dictionary is passed, that overrides
    this default setting.

    You can also configure this device to broadcast readings without a connection on
    a periodic interval as well.

    If no 'trace' argument is passed the device defaults to tracing the phrase
    'Hello trace world.  ' every second.  If a 'trace' array is passed, that overrides
    the default behavior.

    Args:
        args (dict): Any arguments that you want to pass to create this device.
            Supported args are:

                iotile_id (int): The UUID used for this device.  If no UUID is
                    specified, the default value of 1 is used.

                streams (dict): A map of strings with hex numbers to tuples of
                    (interval, value) where interval is a float that expresses
                    how often the stream should stream in seconds and value is
                    an integer that is sent as the value every interval as a
                    realtime reading (IndividualReadingReport).  The stream id
                    is the key of the streams dict which should be a string
                    encoding of a hex number including the prefix 0x so that
                    it can be parsed with int(key, 0).

                broadcast (dict): A map of strings with hex numbers to tuples
                    of (interval, value) where interval is a float that
                    expresses how often the stream should stream in seconds
                    and value is an integer that is sent as the value every
                    interval as a broadcast reading (BroadcastReport).  The
                    stream id is the key of the streams dict which should be a
                    string encoding of a hex number including the prefix 0x so
                    that it can be parsed with int(key, 0).

                    Note that a device can only broadcast a single value at
                    once so if you specify multiple broadcast entries, only
                    the last one to be triggered will be visible at any given
                    time.  For this reason, it is not useful to have multiple
                    broadcast values with the same ``interval`` since only one
                    will ever be shown.

                trace (list): A list of tuples which are (float, string) lists
                    that will trace the fixed string every fixed interval
                    given by the first float argument in seconds.
    """

    def __init__(self, args):
        iotile_id = args.get('iotile_id', 1)

        if isinstance(iotile_id, str):
            iotile_id = int(iotile_id, 16)

        super(RealtimeTestDevice, self).__init__(iotile_id, 'Simple')

        streams = args.get('streams', {'0x1001': (1.0, 100)})
        broadcast = args.get('broadcast', {'0x1001': (1.0, 100)})

        if 'streams' in args:
            streams = args['streams']

        for key, value in streams.items():
            stream = int(key, 0)
            interval, reading = value

            self.create_worker(self._create_stream, interval, stream, reading)

        for key, value in broadcast.items():
            stream = int(key, 0)
            interval, reading = value

            self.create_worker(self._create_broadcast, interval, stream, reading)

        traces = [[1.0, 'Hello trace world.  ']]

        if 'trace' in args:
            traces = args['trace']

        for interval, value in traces:
            self.create_worker(self._create_trace, interval, value)

    async def _create_trace(self, value):
        """Send a realtime tracing value

        Args:
            value (string): The tracing value to send
        """

        if not self.interface_open('tracing'):
            return

        try:
            await self.trace(bytearray(value.encode('ascii')))
        except DevicePushError:
            pass

    async def _create_stream(self, stream, value):
        """Send a realtime streaming value

        Args:
            stream (int): The stream id to send
            value (int): The stream value to send
        """

        if not self.interface_open('streaming'):
            return

        reading = IOTileReading(0, stream, value)

        report = IndividualReadingReport.FromReadings(self.iotile_id, [reading])

        try:
            await self.stream(report)
        except DevicePushError:
            pass

    async def _create_broadcast(self, stream, value):
        """Send a broadcast streaming value.

        Args:
            stream (int): The stream id to send
            value (int): The stream value to send
        """

        reading = IOTileReading(0, stream, value)

        report = BroadcastReport.FromReadings(self.iotile_id, [reading])

        try:
            await self.stream(report)
        except DevicePushError:
            pass



import time
import pytest

from iotile.core.hw.hwmanager import HardwareManager
from iotile.core.exceptions import HardwareError


def pytest_addoption(parser):
    parser.addoption('--port', default="bled112", help="Port to use to connect to iotile device")
    parser.addoption('--uuid', action="append", default=[], help="Device UUIDs to run tests on")
    parser.addoption('--tile', action="append", default=[], help="Device tile addresses to run tests on")
    parser.addoption('--direct', action="store_true", help="Run script on device given by this connection string")
    parser.addoption('--record', help="Record all RPCs to a csv file")


def pytest_generate_tests(metafunc):
    if 'port' in metafunc.fixturenames:
        port = metafunc.config.option.port
        metafunc.parametrize("port", [port], scope='session')

    if 'device_id' in metafunc.fixturenames:
        ids = [int(x, 0) for x in metafunc.config.option.uuid]
        metafunc.parametrize("device_id", ids, scope='session')

    if 'tile' in metafunc.fixturenames:
        ids = [int(x, 0) for x in metafunc.config.option.tile]
        metafunc.parametrize("tile", ids, scope='session')

    if 'direct' in metafunc.fixturenames:
        connections = [metafunc.config.option.direct]
        metafunc.parametrize("direct", connections, scope='session')

    if 'record' in metafunc.fixturenames:
        metafunc.parametrize("record", [metafunc.config.option.record], scope='session')


@pytest.fixture(scope='module')
def device(port, device_id, direct, record, request):
    """Return a HardwareManager instance connected to an IOTile Device

    This fixture shares the same device among all tests in the same module
    """

    if record is not None:
        record = _build_record_string(record, request, module=True)

    with HardwareManager(port=port, record=record) as hw:
        # Sometimes in congested wireless environments we can miss the
        # device advertisement, so attempt the connection several times
        # before giving up to improve test robustness.
        max_attempts = 3
        i = 0
        for i in range(0, max_attempts):
            try:
                if direct:
                    hw.connect_direct(direct)
                else:
                    hw.connect(device_id)
                break
            except HardwareError:
                time.sleep(1)

        if i == max_attempts:
            pytest.fail("Could not connect to device after %d attempts, failing" % max_attempts)
            return

        hw.enable_streaming()
        yield hw
        hw.disconnect()


@pytest.fixture(scope='function')
def per_test_device(port, device_id, direct, record, request):
    """Return a HardwareManager instance connected to an IOTile Device

    This fixture creates and tears down the connection for each test
    """

    if record is not None:
        record = _build_record_string(record, request, module=False)

    with HardwareManager(port=port, record=record) as hw:
        max_attempts = 3
        i = 0
        for i in range(0, max_attempts):
            try:
                if direct:
                    hw.connect_direct(direct)
                else:
                    hw.connect(device_id)
                break
            except HardwareError:
                time.sleep(1)

        if i == max_attempts:
            pytest.fail("Could not connect to device after %d attempts, failing" % max_attempts)
            return

        hw.enable_streaming()
        yield hw
        hw.disconnect()


def _build_record_string(name, request, module=True):
    """Build a proper string name to record connection information."""

    if '{}' not in name:
        return name

    if module:
        identifier = request.module.__name__
    else:
        identifier = request.node.originalname

    return name.format(identifier)
